
MULTIMAP_PUT = 0x0201
MULTIMAP_GET = 0x0202
MULTIMAP_REMOVE = 0x0203
MULTIMAP_KEYSET = 0x0204
MULTIMAP_VALUES = 0x0205
MULTIMAP_ENTRYSET = 0x0206
MULTIMAP_CONTAINSKEY = 0x0207
MULTIMAP_CONTAINSVALUE = 0x0208
MULTIMAP_CONTAINSENTRY = 0x0209
MULTIMAP_SIZE = 0x020a
MULTIMAP_CLEAR = 0x020b
MULTIMAP_VALUECOUNT = 0x020c
MULTIMAP_ADDENTRYLISTENERTOKEY = 0x020d
MULTIMAP_ADDENTRYLISTENER = 0x020e
MULTIMAP_REMOVEENTRYLISTENER = 0x020f
MULTIMAP_LOCK = 0x0210
MULTIMAP_TRYLOCK = 0x0211
MULTIMAP_ISLOCKED = 0x0212
MULTIMAP_UNLOCK = 0x0213
MULTIMAP_FORCEUNLOCK = 0x0214
MULTIMAP_REMOVEENTRY = 0x0215
