
EXECUTORSERVICE_SHUTDOWN = 0x0901
EXECUTORSERVICE_ISSHUTDOWN = 0x0902
EXECUTORSERVICE_CANCELONPARTITION = 0x0903
EXECUTORSERVICE_CANCELONADDRESS = 0x0904
EXECUTORSERVICE_SUBMITTOPARTITION = 0x0905
EXECUTORSERVICE_SUBMITTOADDRESS = 0x0906
