"""
Event Response Constants
"""

EVENT_MEMBER = 200
EVENT_MEMBERLIST = 201
EVENT_MEMBERATTRIBUTECHANGE = 202
EVENT_ENTRY = 203
EVENT_ITEM = 204
EVENT_TOPIC = 205
EVENT_PARTITIONLOST = 206
EVENT_DISTRIBUTEDOBJECT = 207
EVENT_CACHEINVALIDATION = 208
EVENT_MAPPARTITIONLOST = 209
EVENT_CACHE = 210
EVENT_CACHEBATCHINVALIDATION = 211
# ENTERPRISE
EVENT_QUERYCACHESINGLE = 212
EVENT_QUERYCACHEBATCH = 213

EVENT_CACHEPARTITIONLOST = 214
EVENT_IMAPINVALIDATION = 215
EVENT_IMAPBATCHINVALIDATION = 216
