"""
Each exception that are defined in client protocol have unique identifier which are error code.
All error codes defined in protocol are listed in this class.
"""

# Message type indicating exception
EXCEPTION_MESSAGE_TYPE = 109

UNDEFINED = 0
ARRAY_INDEX_OUT_OF_BOUNDS = 1
ARRAY_STORE = 2
AUTHENTICATION = 3
CACHE = 4
CACHE_LOADER = 5
CACHE_NOT_EXISTS = 6
CACHE_WRITER = 7
CALLER_NOT_MEMBER = 8
CANCELLATION = 9
CLASS_CAST = 10
CLASS_NOT_FOUND = 11
CONCURRENT_MODIFICATION = 12
CONFIG_MISMATCH = 13
CONFIGURATION = 14
DISTRIBUTED_OBJECT_DESTROYED = 15
DUPLICATE_INSTANCE_NAME = 16
EOF = 17
ENTRY_PROCESSOR = 18
EXECUTION = 19
HAZELCAST = 20
HAZELCAST_INSTANCE_NOT_ACTIVE = 21
HAZELCAST_OVERLOAD = 22
HAZELCAST_SERIALIZATION = 23
IO = 24
ILLEGAL_ARGUMENT = 25
ILLEGAL_ACCESS_EXCEPTION = 26
ILLEGAL_ACCESS_ERROR = 27
ILLEGAL_MONITOR_STATE = 28
ILLEGAL_STATE = 29
ILLEGAL_THREAD_STATE = 30
INDEX_OUT_OF_BOUNDS = 31
INTERRUPTED = 32
INVALID_ADDRESS = 33
INVALID_CONFIGURATION = 34
MEMBER_LEFT = 35
NEGATIVE_ARRAY_SIZE = 36
NO_SUCH_ELEMENT = 37
NOT_SERIALIZABLE = 38
NULL_POINTER = 39
OPERATION_TIMEOUT = 40
PARTITION_MIGRATING = 41
QUERY = 42
QUERY_RESULT_SIZE_EXCEEDED = 43
QUORUM = 44
REACHED_MAX_SIZE = 45
REJECTED_EXECUTION = 46
REMOTE_MAP_REDUCE = 47
RESPONSE_ALREADY_SENT = 48
RETRYABLE_HAZELCAST = 49
RETRYABLE_IO = 50
RUNTIME = 51
SECURITY = 52
SOCKET = 53
STALE_SEQUENCE = 54
TARGET_DISCONNECTED = 55
TARGET_NOT_MEMBER = 56
TIMEOUT = 57
TOPIC_OVERLOAD = 58
TOPOLOGY_CHANGED = 59
TRANSACTION = 60
TRANSACTION_NOT_ACTIVE = 61
TRANSACTION_TIMED_OUT = 62
URI_SYNTAX = 63
UTF_DATA_FORMAT = 64
UNSUPPORTED_OPERATION = 65
WRONG_TARGET = 66
XA = 67
ACCESS_CONTROL = 68
LOGIN = 69
UNSUPPORTED_CALLBACK = 70
NO_DATA_MEMBER = 71
REPLICATED_MAP_CANT_BE_CREATED = 72
MAX_MESSAGE_SIZE_EXCEEDED = 73
WAN_REPLICATION_QUEUE_FULL = 74
ASSERTION_ERROR = 75
OUT_OF_MEMORY_ERROR = 76
STACK_OVERFLOW_ERROR = 77
NATIVE_OUT_OF_MEMORY_ERROR = 78
SERVICE_NOT_FOUND = 79
STALE_TASK_ID = 80
DUPLICATE_TASK = 81
STALE_TASK = 82
LOCAL_MEMBER_RESET = 83
INDETERMINATE_OPERATION_STATE = 84
FLAKE_ID_NODE_ID_OUT_OF_RANGE_EXCEPTION = 85
TARGET_NOT_REPLICA_EXCEPTION = 86
MUTATION_DISALLOWED_EXCEPTION = 87
CONSISTENCY_LOST_EXCEPTION = 88
