
SET_SIZE = 0x0601
SET_CONTAINS = 0x0602
SET_CONTAINSALL = 0x0603
SET_ADD = 0x0604
SET_REMOVE = 0x0605
SET_ADDALL = 0x0606
SET_COMPAREANDREMOVEALL = 0x0607
SET_COMPAREANDRETAINALL = 0x0608
SET_CLEAR = 0x0609
SET_GETALL = 0x060a
SET_ADDLISTENER = 0x060b
SET_REMOVELISTENER = 0x060c
SET_ISEMPTY = 0x060d
