
MAP_PUT = 0x0101
MAP_GET = 0x0102
MAP_REMOVE = 0x0103
MAP_REPLACE = 0x0104
MAP_REPLACEIFSAME = 0x0105
MAP_CONTAINSKEY = 0x0109
MAP_CONTAINSVALUE = 0x010a
MAP_REMOVEIFSAME = 0x010b
MAP_DELETE = 0x010c
MAP_FLUSH = 0x010d
MAP_TRYREMOVE = 0x010e
MAP_TRYPUT = 0x010f
MAP_PUTTRANSIENT = 0x0110
MAP_PUTIFABSENT = 0x0111
MAP_SET = 0x0112
MAP_LOCK = 0x0113
MAP_TRYLOCK = 0x0114
MAP_ISLOCKED = 0x0115
MAP_UNLOCK = 0x0116
MAP_ADDINTERCEPTOR = 0x0117
MAP_REMOVEINTERCEPTOR = 0x0118
MAP_ADDENTRYLISTENERTOKEYWITHPREDICATE = 0x0119
MAP_ADDENTRYLISTENERWITHPREDICATE = 0x011a
MAP_ADDENTRYLISTENERTOKEY = 0x011b
MAP_ADDENTRYLISTENER = 0x011c
MAP_ADDNEARCACHEENTRYLISTENER = 0x011d
MAP_REMOVEENTRYLISTENER = 0x011e
MAP_ADDPARTITIONLOSTLISTENER = 0x011f
MAP_REMOVEPARTITIONLOSTLISTENER = 0x0120
MAP_GETENTRYVIEW = 0x0121
MAP_EVICT = 0x0122
MAP_EVICTALL = 0x0123
MAP_LOADALL = 0x0124
MAP_LOADGIVENKEYS = 0x0125
MAP_KEYSET = 0x0126
MAP_GETALL = 0x0127
MAP_VALUES = 0x0128
MAP_ENTRYSET = 0x0129
MAP_KEYSETWITHPREDICATE = 0x012a
MAP_VALUESWITHPREDICATE = 0x012b
MAP_ENTRIESWITHPREDICATE = 0x012c
MAP_ADDINDEX = 0x012d
MAP_SIZE = 0x012e
MAP_ISEMPTY = 0x012f
MAP_PUTALL = 0x0130
MAP_CLEAR = 0x0131
MAP_EXECUTEONKEY = 0x0132
MAP_SUBMITTOKEY = 0x0133
MAP_EXECUTEONALLKEYS = 0x0134
MAP_EXECUTEWITHPREDICATE = 0x0135
MAP_EXECUTEONKEYS = 0x0136
MAP_FORCEUNLOCK = 0x0137
MAP_KEYSETWITHPAGINGPREDICATE = 0x0138
MAP_VALUESWITHPAGINGPREDICATE = 0x0139
MAP_ENTRIESWITHPAGINGPREDICATE = 0x013a
MAP_CLEARNEARCACHE = 0x013b
